#include <iostream>
#include <sstream>
#include <ctime>
#include <map>
#include <vector>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <openssl/md5.h>
#include "TomiTCP/net.h"
using namespace std;

#define LISTEN_PORT 6667
/* 
 * note: I strongly recommend having IPs here, or having hostnames in
 * /etc/hosts, or having 100% reliable named
 */
#define SD_HOST "127.0.0.1"
#define SD_PORT 6666
#define HYB_HOST "62.77.119.154"
#define HYB_PORT 6667
#define REALLEN 50
#define NICKLEN 20
#define SERVNICKLEN 14

net::TomiTCP *hyb, *sd;

map<int,string> servers;
// servers[0] = hybrid server name, servers[1] = SDr server name
int sinc = 2;

map<string,string> den_chan;

bool den_ch(string chan)
{
    if (den_chan.find(chan) != den_chan.end())
	return true;

    return false;
}

// ------ code for generating service nick
#define CHIDNB 36
#define CHIDLEN SERVNICKLEN

string sernick = "";

static unsigned char id_alphabet[CHIDNB+1] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";

char * ltoid(unsigned int l)
{
    static char idrpl[CHIDLEN+1];
    char c = CHIDLEN-1;

    idrpl[CHIDLEN] = '\0';
    do {
	idrpl[(unsigned int)c] = id_alphabet[l % CHIDNB];
	l /= CHIDNB;
    } while (c-- > 0);

    return (char *) idrpl;
}

char md5me[CHIDLEN+1];
void    md_server_name(const char *a, const char *b)
{
    unsigned char    mdbuf[16];
    MD5_CTX md5;
    MD5_Init(&md5);
    if (a) MD5_Update(&md5, a, strlen(a));
    if (b) MD5_Update(&md5, b, strlen(b));
    MD5_Final(mdbuf, &md5);
    memcpy(md5me, ltoid(*((time_t *) mdbuf)), CHIDLEN);
    md5me[CHIDLEN] = 0;
}

void make_servnick()
{
    md_server_name(servers[0].c_str(),servers[1].c_str());
    sernick = string("anonet") + md5me;
}
// ------

// get server token from server name
int sint(string s)
{
    for (map<int,string>::iterator i = servers.begin(); i != servers.end(); i++) {
	if (i->second == s) {
	    if (i->first == 0)
		return 1;
	    return i->first;
	}
    }

    return 1;
}

string intstr(int i)
{
    char s[32];
    sprintf(s,"%i",i);
    return s;
}

// unmangle collision recovered nicks
string sd_nick_to(string n)
{
    if (n.length() < 7)
	return n;

    if (string(n,0,6) == "anonet" && n[6] >= '0' && n[6] <= '9') {
	return string(n,6);
    }

    return n;
}

// mangle collision recovered nicks
string sd_nick_from(string n)
{
    if (n.length() == 9 && n[0] >= '0' && n[0] <= '9') {
	return string("anonet" + n,0,NICKLEN);
    }

    return n;
}

// combine mode array into mode string
string combinemode(vector<string> modes)
{
    string a,b;
    char ch = ' ';

    for (vector<string>::iterator i = modes.begin(); i != modes.end(); i++) {
	string c,d;

	stringstream s(*i);
	s >> c;
	getline(s,d);
	d.erase(0,d.find_first_not_of(" \f\n\r\t\v"));

	if (c.length() > 1) {
	    if (c[0] != ch) {
		a += c;
		ch = c[0];
	    } else {
		a += string(c,1);
	    }

	    if (d.length()) {
		b += " " + d;
	    }
	}
    }

    return a + b;
}

// parse hybrid modes
void hyb_parsemode(string modes, vector<string>& mode)
{
    char ch[2] = "+",t[2] = " ";
    stringstream s(modes);
    string cmd;
    s >> cmd;

    for (string::iterator i=cmd.begin(); i!=cmd.end(); i++) {
	if (*i == '+' || *i == '-')
	    ch[0] = *i;
	else {
	    string m;

	    t[0] = *i;
	    switch (t[0]) {
		case 'b':
		case 'e':
		case 'I':
		case 'k':
		case 'o':
		case 'h':
		case 'v':
		    s >> m;
		    mode.push_back(string(ch)+string(t)+((m.length())?(" "+m):""));
		    break;
		case 'l':
		    if (ch[0] == '+') {
			s >> m;
			mode.push_back(string(ch)+string(t)+" "+m);
		    } else {
			mode.push_back(string(ch)+string(t));
		    }
		    break;
		default:
		    mode.push_back(string(ch)+string(t));
	    }
	}
    }
}

// do something with hybrid modes
string hyb_chan_modes(string m, string chan)
{
    vector<string> mode, modes;
    hyb_parsemode(m,mode);

    for (vector<string>::iterator i = mode.begin(); i != mode.end(); i++) {
	if (i->length() > 1) {
	    switch ((*i)[1]) {
		case 'o':
		case 'h':
		case 'v':
		    if (i->length() > 3) {
			modes.push_back(string(*i,0,3) + sd_nick_to(string(*i,3)));
		    } else
			modes.push_back(*i);
		    break;
		default:
		    modes.push_back(*i);
	    }
	}
    }

    m = combinemode(modes);

    return m;
}

// work on data from hybrid...
string hyb_in(string m)
{
    stringstream s(m);

    string cmd, src;
    s >> cmd;
    if (cmd.length() && cmd[0] == ':') {
	src = cmd;
	s >> cmd;
    }

    if (cmd == "PASS") {
	string pass;
	s >> pass;

	return cmd + " " + pass + " 0210030003 IRC|QaCDeEfHiIjJKMprRstTwZ6@ P\n";
    }

    if (cmd == "SERVER") {
	if (src.length()) {
	    string name, hop, comment;
	    s >> name >> hop;
	    getline(s,comment);

	    comment.erase(0,comment.find_first_not_of(" \f\n\r\t\v"));
	    comment.erase(comment.find_last_not_of(" \f\n\r\t\v")+1);
	    if (comment.length() && comment[0] == ':') {
		comment.erase(comment.begin());
	    }

	    int si = sint(name);
	    if (si == 1) {
		si = sinc;
		sinc++;
		servers[si] = name;
	    }
	    string tok = intstr(si);
	    return src + " " + cmd + " " + name + " " + hop + " " + tok + " :" + comment + "\n";
	} else {
	    string name;
	    s >> name;
	    servers[0] = name;
	    make_servnick();
	}
	return m;
    }

    if (cmd == "NICK") {
	if (src.length()) {
	    string nick;
	    s >> nick;

	    if (nick == sernick) {
		fprintf(*hyb,"KILL %s :%s Nick is reserved for gateway\n",nick.c_str(),
			servers[1].c_str());
		return "";
	    }

	    return src + " " + cmd + " " + nick + "\n";
	}

	string nick, hop, ts, mode, user, host, server, real;
	s >> nick >> hop >> ts >> mode >> user >> host >> server;
	getline(s,real);

	real.erase(0,real.find_first_not_of(" \f\n\r\t\v"));
	real.erase(real.find_last_not_of(" \f\n\r\t\v")+1);
	if (real.length() && real[0] == ':') {
	    real.erase(real.begin());
	}
	if (real.length() > REALLEN)
	    real = string(real,0,REALLEN);

	for (string::iterator i = mode.begin(); i != mode.end(); i++) {
	    if (*i == 'a') {
		mode.erase(i);
		break;
	    }
	}

	if (nick == sernick) {
	    fprintf(*hyb,"KILL %s :%s Nick is reserved for gateway\n",nick.c_str(),
		    servers[1].c_str());
	    return "";
	}

	return cmd + " " + nick + " " + hop + " " + user + " " + host + " " +
	    intstr(sint(server)) + " " + mode + " :" + real + "\n";
    }

    if (cmd == "SJOIN") {
	string a,chan,mode,list;
	s >> a >> chan;
	getline(s,mode,':');
	mode.erase(0,mode.find_first_not_of(" \f\n\r\t\v"));
	mode.erase(mode.find_last_not_of(" \f\n\r\t\v")+1);
	while (getline(s,a,' ')) {
	    a.erase(a.find_last_not_of(" \f\n\r\t\v")+1);
	    if (!a.length()) continue;
	    if (list.length())
		list += ",";
	    list += a;
	}

	if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
	    return "";

	if (den_ch(chan)) {
	    string klist,nlist;
	    stringstream sl(list);
	    while (getline(sl,a,',')) {
		if (a.length() && a[0] == '@') {
		    if (!a.length()) continue;
		    if (nlist.length())
			nlist += ",";
		    nlist += a;
		} else {
		    a.erase(0,a.find_first_not_of("@+"));
		    if (!a.length()) continue;
		    if (klist.length())
			klist += ",";
		    klist += a;
		}
	    }

	    if (klist.length()) {
		fprintf(*hyb,"NICK %s 1 1 +oi services services.int %s :Gateway service\n",
			sernick.c_str(), servers[1].c_str());
		fprintf(*hyb,":%s SJOIN 0 %s + :@%s\n",servers[1].c_str(), chan.c_str(),
			sernick.c_str());
		fprintf(*hyb,":%s KICK %s %s :cannot join +S/+c channel from hybrid side\n",
			sernick.c_str(), chan.c_str(), klist.c_str());
		fprintf(*hyb,":%s QUIT :Have a nice day\n",sernick.c_str());
	    }
	    if (nlist.length())
		return src + " NJOIN " + chan + " :" + nlist + "\n" + src + " MODE " + chan + " " + mode + "\n";
	    else
		return "";
	}

	return src + " NJOIN " + chan + " :" + list + "\n" + src + " MODE " + chan + " " + mode + "\n";
    }

    if (cmd == "MODE") {
	string target, modes;
	s >> target;
	getline(s,modes);
	
	modes.erase(0,modes.find_first_not_of(" \f\n\r\t\v"));
	modes.erase(modes.find_last_not_of(" \f\n\r\t\v")+1);
	if (modes.length() && modes[0] == ':') {
	    modes.erase(modes.begin());
	}

	if (target.length()) {
	    if (target.length() && (target[0] == '!' || target[0] == '+'))
		return "";
	    if (target[0] != '#') {
		for (string::iterator i = modes.begin(); i != modes.end(); i++) {
		    if (*i == 'a') {
			modes.erase(i);
			break;
		    }
		}
	    } else {
		modes = hyb_chan_modes(modes,target);
	    }

	    return ((src.length())?(src+" "):"") + cmd + " " + target + " " + modes + "\n";
	}
    }

    if (cmd == "PART") {
	string chans,res,nchans;
	s >> chans;
	getline(s,res);
	string modes;

	stringstream s2(chans);
	string chan;
	while (getline(s2,chan,',')) {
	    if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
		continue;

	    if (nchans.length())
		nchans += ",";
	    nchans += chan;
	    modes += ":" + servers[0] + " MODE " + chan + "\n";
	}
	return ((src.length())?(src+" "):"") + cmd + " " + nchans + res + "\n" + modes;
    }

    if (cmd == "PRIVMSG" || cmd == "NOTICE") {
	string chan,res;
	s >> chan;
	getline(s,res);

	if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
	    return "";
	if (chan.length() && chan[0] != '#') {
	    chan = sd_nick_to(chan);
	}

	if (cmd == "NOTICE" && chan == "AUTH")
	    return "";

	return ((src.length())?(src+" "):"") + cmd + " " + chan + res + "\n";
    }

    if (cmd == "KICK") {
	string chans,nicks,nnicks,nick,com;
	s >> chans >> nicks;
	getline(s,com);

	stringstream n(nicks);
	while (getline(n,nick,',')) {
	    if (nnicks.length())
		nnicks += ",";
	    nnicks += sd_nick_to(nick);
	}

	return ((src.length())?(src+" "):"") + cmd + " " + chans + " " + nnicks + com + "\n";
    }

    if (cmd == "OPERWALL") {
	string msg;
	getline(s,msg);

	msg.erase(0,msg.find_first_not_of(" \f\n\r\t\v"));
	msg.erase(msg.find_last_not_of(" \f\n\r\t\v")+1);
	if (msg.length() && msg[0] == ':') {
	    msg.erase(msg.begin());
	}

	return ((src.length())?(src+" "):"") + "WALLOPS :OPERWALL - " + msg + "\n";
    }

    if (cmd == "CAPAB")
	return "";
    if (cmd == "SVINFO")
	return "";
    if (cmd == "KLINE")
	return "";
    if (cmd == "FORCEJOIN" || cmd == "FORCEPART")
	return ""; // we can warn user that it will not success

    return m;
}

// make nick TS
string TS()
{
    /*time_t t = time(0);
    char s[32];
    sprintf(s,"%i",(int)t);
    return s;*/
    return "2"; // Note: announce nicks with TS 2 and service nicks with TS 1
}

// parse SDr modes
void sd_parsemode(string modes, vector<string>& mode)
{
    char ch[2] = "+",t[2] = " ";
    stringstream s(modes);
    string cmd;
    s >> cmd;

    for (string::iterator i=cmd.begin(); i!=cmd.end(); i++) {
	if (*i == '+' || *i == '-')
	    ch[0] = *i;
	else {
	    string m;

	    t[0] = *i;
	    switch (t[0]) {
		case 'b':
		case 'e':
		case 'I':
		case 'C':
		case 'F':
		case 'k':
		case 'o':
		case 'h':
		case 'v':
		    s >> m;
		    mode.push_back(string(ch)+string(t)+((m.length())?(" "+m):""));
		    break;
		case 'l':
		    if (ch[0] == '+') {
			s >> m;
			mode.push_back(string(ch)+string(t)+" "+m);
		    } else {
			mode.push_back(string(ch)+string(t));
		    }
		    break;
		default:
		    mode.push_back(string(ch)+string(t));
	    }
	}
    }
}

// do something with SDr modes (mainly stripping)
string sd_chan_modes(string m, string chan)
{
    vector<string> mode, modes;
    sd_parsemode(m,mode);

    for (vector<string>::iterator i = mode.begin(); i != mode.end(); i++) {
	if (i->length() > 1) {
	    switch ((*i)[1]) {
		case 'S':
		case 'c':
		    if ((*i)[0] == '+') {
			if (den_chan[chan].find((*i)[1]) == string::npos)
			    den_chan[chan] += (*i)[1];
		    } else {
			for (string::iterator j = den_chan[chan].begin(); j != den_chan[chan].end(); j++) {
			    if (*j == (*i)[1]) {
				den_chan[chan].erase(j);
				break;
			    }
			}
			if (!den_chan[chan].length())
			    den_chan.erase(chan);
		    }
		    break;
		case 'q':
		case 'C':
		case 'K':
		case 'F':
		case 'O':
		case 'a':
		case 'r':
		    break;
		case 'o':
		case 'h':
		case 'v':
		    if (i->length() > 3) {
			modes.push_back(string(*i,0,3) + sd_nick_from(string(*i,3)));
		    } else
			modes.push_back(*i);
		    break;
		default:
		    modes.push_back(*i);
	    }
	}
    }

    m = combinemode(modes);

    return m;
}

// do somthing with SDr nick modes (away...)
string sd_nick_modes(string m, int &away)
{
    char ch = '+';
    string n;

    away = 0;

    for (string::iterator i = m.begin(); i != m.end(); i++) {
	if (*i == '+' || *i == '-') {
	    ch = *i;
	    if (n.length() && (*(n.end()-1) == '+' || *(n.end()-1) == '-'))
		*(n.end()-1) = ch;
	    else
		n += ch;
	} else {
	    if (*i == 'a') {
		if (ch == '+') {
		    away = 1;
		} else {
		    away = 2;
		}
	    }
	    if (*i == 'F' || *i == 'I' || *i == 'a')
		continue;

	    n += *i;
	}
    }

    if (n == "+" || n == "-")
	return "";

    return n;
}

// work on data from SDr
string sd_in(string m)
{
    stringstream s(m);

    string cmd, src;
    s >> cmd;
    if (cmd.length() && cmd[0] == ':') {
	src = cmd;
	s >> cmd;
    }
    if (src.length()) {
	string newsrc = ":" + sd_nick_from(string(src,1));
	if (newsrc.length() != src.length()) {
	    m.replace(0,src.length(),newsrc);
	    src = newsrc;
	}
    }

    if (cmd == "PASS") {
	string pass;
	s >> pass;

	return cmd + " " + pass + " :TS\n" + "CAPAB :TS EX IE HUB\n";
    }

    if (cmd == "NICK") {
	if (src.length()) {
	    string nick;
	    s >> nick;
	    if (nick.length() && nick[0] == ':') {
		nick.erase(nick.begin());
	    }

	    string old = src;
	    if (old.length() && old[0] == ':') {
		old.erase(old.begin());
	    }
	    old = sd_nick_from(old);

	    if (nick == sernick) {
		fprintf(*sd,"KILL %s :%s Nick is reserved for gateway\n",nick.c_str(),
			servers[0].c_str());
		return "";
	    }

	    return ":" + old + " " + cmd + " " + nick + " :" + TS() + "\n";
	}

	string nick, hop, user, host, stok, mode, real;
	s >> nick >> hop >> user >> host >> stok >> mode;
	getline(s,real);

	nick = sd_nick_from(nick);

	real.erase(0,real.find_first_not_of(" \f\n\r\t\v"));
	real.erase(real.find_last_not_of(" \f\n\r\t\v")+1);
	if (real.length() && real[0] == ':') {
	    real.erase(real.begin());
	}
	if (real.length() > REALLEN)
	    real = string(real,0,REALLEN);

	string awaym; int away;
	mode = sd_nick_modes(mode,away);
	if (away) {
	    if (away == 1)
		awaym = ":" + nick + " AWAY :Gone";
	    else
		awaym = ":" + nick + " AWAY";
	}
	if (!mode.length())
	    mode = "+";

	if (nick == sernick) {
	    fprintf(*sd,"KILL %s :%s Nick is reserved for gateway\n",nick.c_str(),
		    servers[0].c_str());
	    return "";
	}

	return cmd + " " + nick + " " + hop + " " + TS() + " " + mode +
	    " " + user + " " + host + " " + servers[atol(stok.c_str())] +
	    " :" + real + "\n" + ((awaym.length())?(awaym+"\n"):"");
    }

    if (cmd == "SERVER") {
	if (src.length()) {
	    string name, hop, comment, tok;
	    s >> name >> hop >> tok;
	    getline(s,comment);

	    comment.erase(0,comment.find_first_not_of(" \f\n\r\t\v"));
	    comment.erase(comment.find_last_not_of(" \f\n\r\t\v")+1);
	    if (comment.length() && comment[0] == ':') {
		comment.erase(comment.begin());
	    }

	    int itok = atol(tok.c_str());
	    servers[itok] = name;
	    if (sinc <= itok)
		sinc = itok + 1;

	    return src + " " + cmd + " " + name + " " + hop + " :" + comment + "\n";
	} else {
	    string name;
	    s >> name;
	    servers[1] = name;
	    make_servnick();
	}
	return m;
    }

    if (cmd == "JOIN") {
	string a,chan,mode,nick,smode,b;
	getline(s,b);
	b.erase(0,b.find_first_not_of(" :"));
	stringstream s1(b);

	string out;
	while (getline(s1,a,',')) {
	    stringstream s2(a);
	    smode = "";

	    getline(s2,chan,'');
	    chan.erase(chan.find_last_not_of(" \f\n\r\t\v")+1);
	    getline(s2,mode);

	    if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
		continue;


	    nick = src;
	    if (nick.length() && nick[0] == ':') {
		nick.erase(nick.begin());
	    }
	    nick = sd_nick_from(nick);

	    for (string::iterator i = mode.begin(); i != mode.end(); i++) {
		if (*i == 'O' || *i == 'o')
		    smode += "@";
		if (*i == 'v')
		    smode += "+";
	    }

	    out += ":" + servers[1] + " SJOIN 0 " + chan + " + :" + smode + nick + "\n";
	}

	return out;
    }

    if (cmd == "NJOIN") {
	string a,chan,mode,list;
	s >> chan;

	if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
	    return "";

	getline(s,a,':');
	while (getline(s,a,',')) {
	    a.erase(a.find_last_not_of(" \f\n\r\t\v")+1);
	    if (!a.length()) continue;
	    if (list.length())
		list += " ";
	    list += sd_nick_from(a);
	}

	return src + " SJOIN 0 " + chan + " + :" + list + "\n";
    }

    if (cmd == "MODE") {
	string target, modes;
	s >> target;
	getline(s,modes);
	
	modes.erase(0,modes.find_first_not_of(" \f\n\r\t\v"));
	modes.erase(modes.find_last_not_of(" \f\n\r\t\v")+1);
	if (modes.length() && modes[0] == ':') {
	    modes.erase(modes.begin());
	}

	if (target.length()) {
	    if (target[0] == '!' || target[0] == '+')
		return "";

	    string awaym;

	    if (target[0] == '#') {
		modes = sd_chan_modes(modes,target);
	    } else {
		target = sd_nick_from(target);

		int away;
		modes = sd_nick_modes(modes,away);
		if (away) {
		    if (away == 1)
			awaym = ":" + target + " AWAY :Gone";
		    else
			awaym = ":" + target + " AWAY";
		}
	    }

	    if (!modes.length())
		return "" + ((awaym.length())?(awaym+"\n"):"");

	    return ((src.length())?(src+" "):"") + cmd + " " + target + " " + modes + "\n" +
		((awaym.length())?(awaym+"\n"):"");
	}
    }

    if (cmd == "PART") {
	string chans,res,nchans;
	s >> chans;
	getline(s,res);

	stringstream s2(chans);
	string chan;
	while (getline(s2,chan,',')) {
	    if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
		continue;

	    if (nchans.length())
		nchans += ",";
	    nchans += chan;

	    fprintf(*sd,":%s MODE %s\n",servers[0].c_str(),chan.c_str());
	}

	if (!nchans.length())
	    return "";

	return ((src.length())?(src+" "):"") + cmd + " " + nchans + res + "\n";
    }

    if (cmd == "PRIVMSG" || cmd == "NOTICE") {
	string chan,res;
	s >> chan;
	getline(s,res);

	if (chan.length() && (chan[0] == '!' || chan[0] == '+'))
	    return "";
	return ((src.length())?(src+" "):"") + cmd + " " + chan + res + "\n";
    }

    if (cmd == "KICK") {
	string chans,nicks,nnicks,nick,com;
	s >> chans >> nicks;
	getline(s,com);

	stringstream n(nicks);
	while (getline(n,nick,',')) {
	    if (nnicks.length())
		nnicks += ",";
	    nnicks += sd_nick_from(nick);
	}

	return ((src.length())?(src+" "):"") + cmd + " " + chans + " " + nnicks + com + "\n";
    }

    if (cmd == "NAMES")
	return "";
    if (cmd == "LIST")
	return "";
    if (cmd == "403") {
	string t;
	s >> t;

	if (t == servers[0])
	    return "";
	return m;
    }
    if (cmd == "324") {
	string t;
	s >> t;

	if (t == servers[0]) {
	    string chan,mode;
	    s >> chan;
	    getline(s,mode);
	    mode.erase(0,mode.find_first_not_of(" \f\n\r\t\v"));

	    map<string,string>::iterator i = den_chan.find(chan);
	    if (i != den_chan.end())
		den_chan.erase(i);

	    sd_chan_modes(mode,chan);
	    return "";
	}
	return m;
    }

    return m;
}

int main(int argc, char *argv[])
{
    string buf;

    signal(SIGPIPE,SIG_IGN);

    try {
	net::TomiTCP ls(LISTEN_PORT);

	while (1) {
	    hyb = 0; sd = 0;
	    sinc = 2; servers.clear(); den_chan.clear();

	    try {
		net::TomiTCP *n = ls.accept();
		string host = tomi_ntop(n->rname);
		string host2 = revers(n->rname);
		if (host == HYB_HOST || host2 == HYB_HOST) {
		    hyb = n;
		    try {
			sd = new net::TomiTCP(SD_HOST, SD_PORT);
		    } catch (runtime_error e) {
			fprintf(*n,"ERROR :(gate) Closing link: Connect problem (%s)\n",
				e.what());
			throw;
		    }
		} else if (host == SD_HOST || host2 == SD_HOST) {
		    sd = n;
		    try {
			hyb = new net::TomiTCP(HYB_HOST, HYB_PORT);
		    } catch (runtime_error e) {
			fprintf(*n,"ERROR :(gate) Closing link: Connect problem (%s)\n",
				e.what());
			throw;
		    }
		} else {
		    fprintf(*n,"ERROR :(gate) Closing link: %s (Unathorized connection)\n",
			    host.c_str());
		    delete n;
		    throw runtime_error("unauthorized connection - "+host);
		}

		cout << "NEW" << endl;

		while (1) {
		    fd_set set;

		    FD_ZERO(&set);
		    FD_SET(hyb->sock, &set);
		    FD_SET(sd->sock, &set);

		    int ret = TEMP_FAILURE_RETRY(
			    select(FD_SETSIZE,&set,NULL,NULL,0));
		    if (ret == -1)
			throw runtime_error("error in select: "+string(strerror(errno)));

		    if (FD_ISSET(hyb->sock, &set)) {
			if (! hyb->getline(buf))
			    throw runtime_error("hyb - getline error (maybe EOF)");

			cout << "<--H " << buf;
			string m = hyb_in(buf);
			if (m.length()) {
			    cout << "<==H " << m;
			    fputs(m.c_str(),*sd);
			}
		    }

		    if (FD_ISSET(sd->sock, &set)) {
			if (! sd->getline(buf))
			    throw runtime_error("sd - getline error (maybe EOF)");

			cout << "S--> " << buf;
			string m = sd_in(buf);
			if (m.length()) {
			    cout << "S==> " << m;
			    fputs(m.c_str(),*hyb);
			}
		    }
		}

	    } catch (runtime_error e) {
		cerr << e.what() << endl;
	    }

	    delete hyb; hyb = 0;
	    delete sd; sd = 0;
	}
    } catch (runtime_error e) {
	cerr << e.what() << endl;
    }
    return 0;
}
